//#include "hardware.h"
//#include "adc.h"

#include <stdint.h>
#include <math.h>
#include "ntc.h"
#include <stdio.h>

//http://www.giangrandi.ch/electronics/ntc/ntc.shtml

//#define RREF 100000L ikea
#define RREF 47000

#define R25	 100000L   /// vishay
//#define R25	 220000L  // ikea
//#define BVALUE 3977 10k
#define BVALUE 4190 // 100k 2322 640 5 Vishay
//#define BVALUE 4300  // Ikea
//#define BVALUE 4225  // Ikea

void sendToApp( char * str);

// scaling:
//1   = 0
//2/3 = 1
//1/3 = 2

// ref :
// 0 = 1/2 supp
// 1 = 1/3 supp

uint32_t NTCaveragingBuffer[NR_NTCS][NTCAVERAGING + 2];

#define ADCMAXVAL 4096  // 12 bits ADC

float temperature[NR_NTCS];
float resistance[NR_NTCS];


extern uint32_t PWMdutyCycle;

float calcTemp(uint32_t ID);
float NTCreferenceValue = 3800;

float calcNTC( uint32_t ADCVal) {
	float ntc;
	if (ADCVal > 4000){
		return (1000); // no ntc
	}


//	ntc = RREF * (float) ADCVal / (ADCMAXVAL) - (float) ADCVal);  // is RREF is connected to ADC reference
	ntc = RREF * (float) ADCVal / ( NTCreferenceValue - (float) ADCVal);  // is RREF is connected to 3V3 with separate reference

	return (ntc);
}

float resToTemp(float Rntc) {
	float temp;
	temp = (1.0 / ((log(Rntc / R25)) / BVALUE +(1.0 / 298.15))) - 273.0; //log = ln
	return (temp);
}

static uint8_t tabPntr[NR_NTCS];

float updateNTC(uint32_t ID, uint32_t adcVal) {
	int temp;
	float r = ( int) calcNTC(adcVal);  // adcVal;

	resistance[ID] = r; // test only
	if (r > 100)
	{
		temp = (int) 100 * resToTemp(r);  // fill buffer with temperatures

		if (NTCaveragingBuffer[ID][0] == 0)
		{
			for (int n = 0; n < NTCAVERAGING; n++)  // fill buffer at startup
				NTCaveragingBuffer[ID][n] = temp;
		}
		else
			NTCaveragingBuffer[ID][tabPntr[ID]] = temp;

		tabPntr[ID]++;
		if (tabPntr[ID] == NTCAVERAGING)
		{
			tabPntr[ID] = 0;
		}
		temperature[ID] = calcTemp(ID);
	}
	else
		temperature[ID] = 9999;


	return temperature[ID];
}

// berekent uit gemiddelde van tabel de temperatuur
// returnwaarde: temperatuur in graden C
// input

float calcTemp(uint32_t ID) {
	uint32_t * pntr;
	uint32_t sum = 0;
	uint8_t n;
	uint32_t max = 0, min = 0xFFFFF;

	float Rntc, temperature;

	pntr = NTCaveragingBuffer[ID];

	if (NTCAVERAGING > 2)
	{
		for (n = 0; n < NTCAVERAGING; n++)
		{
			if (*pntr > max)
				max = *pntr;
			if (*pntr < min)
				min = *pntr;
			sum += *pntr++;
		}
		sum -= max;  // reject highest and lowest
		sum -= min;
		sum /= NTCAVERAGING - 2;
	}
	else
	{
		for (n = 0; n < NTCAVERAGING; n++)
		{
			sum += *pntr++;
		}
		sum /= NTCAVERAGING;
	}
	return ( sum/ 100.0);
}

//
//void temperatureTimeoutHandler(void * p_context) {
//	static uint32_t state = 0;
//	static uint32_t samples;
//	static int waitcntr;
//	static float temp;
//	static uint32_t id = 0;
//	static uint32_t adc_sample;
//	char str[50];
//
//	switch ( state){
//	case 0:
//		setConfigADC(id);
//		nrf_adc_input_select(NRF_ADC_CONFIG_INPUT_2 << id );  //NRF_ADC_CONFIG_INPUT_2 = P0.1
//		nrf_adc_start();
//		state++;
//		break;
//	case 1:
//		if (nrf_adc_conversion_finished()){
//			nrf_adc_conversion_event_clean();
//			adc_sample = nrf_adc_result_get();
//			if (adc_sample >= 1020) {
//				temperature[id] = 9999;
//			}
//			else
//				temp = updateNTC(id, adc_sample);
//
//			setNextRange(id, adc_sample);
//
//
//		//	temp = calcNTC(id, adc_sample);
//		//	SEGGER_RTT_printf(0, "\r\ntemp %d: %d %d\n\r", id+1,adc_sample, (int) (100.0 * temp));
//			sprintf( str,"T%d: %2.2f %3.0f", id+1, temperature[id], resistance[id]);
//
//			SEGGER_RTT_printf(0, " %s", str);
//			sendToApp(str);
//			id++;
//			if ( id >=  NR_NTCS){
//				sprintf( str,"PWM: %d\n\r",PWMdutyCycle);
//				SEGGER_RTT_printf(0, " %s", str);
//				sendToApp(str);
//				id = 0;
//			}
//
//			//SEGGER_RTT_printf(0, "\r\ntemp %d: %d %d\n\r", id+1,adc_sample, (int) (temp));
//
//
//			state = 0;
//		}
//	}
//}

