
#include "freertos/FreeRTOS.h"

#include "esp_log.h"
#include "tcpip_adapter.h"

#include "lwip/err.h"
#include "lwip/sys.h"
#include "lwip/netdb.h"
#include "lwip/api.h"
#include "lwip/mem.h"
#include "lwipopts.h"

#include "lwip/mem.h"
#include "lwipopts.h"
#include "httpserver.h"
#include "cgiScripts.h"

#include "httpd-fs.h"
#include "httpd.h"

//#pragma GCC optimize ("0")

#include <string.h>

#define HTTPD_DEBUG 		LWIP_DBG_OFF
#define INCLUDE_HTTPD_CGI
//#define NETCONS   // else use sockets
#define HTTPDTIMEOUT	5000/portTICK_RATE_MS
#define MAX_TX_SIZE		1024
#define TAG "httpServer"

//#pragma GCC optimize ("0")

extern bool gotIP;
extern bool isConnected;


typedef struct {
	const char *name;
	u8_t shtml;
} default_filename;

const default_filename g_psDefaultFilenames[] = { { "/index.html", true }, {
		"/index.ssi", true }, { "/index.shtm", true },
		{ "/index.shtml", false }, { "/index.htm", false } };

#define NUM_DEFAULT_FILENAMES (sizeof(g_psDefaultFilenames) /         \
                               sizeof(default_filename))

#ifdef INCLUDE_HTTPD_SSI

/* Include the (TI-specific) micro-standard-library.  This is needed for the
 * function usnprintf which is a cut down version of the C standard snprintf.
 * Some of the toolchains we support don't support snprintf without pulling in
 * a large amount of extra (and usually extraneous) code.  If your C runtime
 * does support snprintf, this header can be removed and the call to usnprintf
 * replaced.
 */
#include "ustdlib.h"

const char *g_pcSSIExtensions[] =
{
	".shtml", ".shtm", ".ssi", ".xml"
};

#define NUM_SHTML_EXTENSIONS (sizeof(g_pcSSIExtensions) / sizeof(const char *))

enum tag_check_state
{
	TAG_NONE, /* Not processing an SSI tag */
	TAG_LEADIN, /* Tag lead in "<!--#" being processed */
	TAG_FOUND, /* Tag name being read, looking for lead-out start */
	TAG_LEADOUT, /* Tag lead out "-->" being processed */
	TAG_SENDING /* Sending tag replacement string */
};
#endif /* INCLUDE_HTTPD_SSI */

struct http_state {
	struct fs_file *handle;
	char *file; /* Pointer to first unsent byte in buf. */
	char *buf; /* File read buffer. */
#ifdef INCLUDE_HTTPD_SSI
	char *parsed; /* Pointer to the first unparsed byte in buf. */
	char *tag_end; /* Pointer to char after the closing '>' of the tag. */
	u32_t parse_left; /* Number of unparsed bytes in buf. */
#endif
	u32_t left; /* Number of unsent bytes in buf. */
	int buf_len; /* Size of file read buffer, buf. */
	u8_t retries;
#ifdef INCLUDE_HTTPD_SSI
	u8_t tag_check; /* true if we are processing a .shtml file else false */
	u8_t tag_index; /* Counter used by tag parsing state machine */
	u8_t tag_insert_len; /* Length of insert in string tag_insert */
	u8_t tag_name_len; /* Length of the tag name in string tag_name */
	char tag_name[MAX_TAG_NAME_LEN + 1]; /* Last tag name extracted */
	char tag_insert[MAX_TAG_INSERT_LEN + 1]; /* Insert string for tag_name */
	enum tag_check_state tag_state; /* State of the tag processor */
#endif
#ifdef INCLUDE_HTTPD_CGI
	char *params[MAX_CGI_PARAMETERS]; /* Params extracted from the request URI */
	char *param_vals[MAX_CGI_PARAMETERS]; /* Values for each extracted param */
#endif

};

#ifdef INCLUDE_HTTPD_SSI
/* SSI insert handler function pointer. */
tSSIHandler g_pfnSSIHandler = NULL;
int g_iNumTags = 0;
const char **g_ppcTags = NULL;

#define LEN_TAG_LEAD_IN 5
const char * const g_pcTagLeadIn = "<!--#";

#define LEN_TAG_LEAD_OUT 3
const char * const g_pcTagLeadOut = "-->";
#endif /* INCLUDE_HTTPD_SSI */

#ifdef INCLUDE_HTTPD_CGI
/* CGI handler information */
const tCGI *g_pCGIs = NULL;
int g_iNumCGIs = 0;
#endif /* INCLUDE_HTTPD_CGI */

/*-----------------------------------------------------------------------------------*/
static struct fs_file *
get_404_file(char **ppURI) {
	struct fs_file *file;

	*ppURI = "data_404_htm"; // "/404.html";
	file = fs_open(*ppURI);
	if (file == NULL) {
		/* 404.html doesn't exist. Try 404.htm instead. */
		*ppURI = "/404.htm";
		file = fs_open(*ppURI);
		if (file == NULL) {
			/* 404.htm doesn't exist either. Indicate to the caller that it should
			 * send back a default 404 page.
			 */
			*ppURI = NULL;
		}
	}

	return (file);
}

//static struct http_state * http_accept (struct netconn *conn)
static struct http_state * http_accept(void) {
	struct http_state *hs;
//	LWIP_DEBUGF(HTTPD_DEBUG, ("http_accept 0x%08x\n", (unsigned int ) conn));

	hs = (struct http_state *) mem_malloc(sizeof(struct http_state));

	if (hs == NULL) {
		LWIP_DEBUGF(HTTPD_DEBUG, ("http_accept: Out of memory\n"));
		return NULL;
	}
	/* Initialize the structure. */
	hs->handle = NULL;
	hs->file = NULL;
	hs->buf = NULL;
	hs->buf_len = 0;
	hs->left = 0;
	hs->retries = 0;

	return hs;

}

/*-----------------------------------------------------------------------------------*/
#ifdef NETCONS
static void send_data (struct netconn *pConn,struct http_state *pHs)
#else
static void send_data(int socket, struct http_state *pHs)
#endif
{
	int count;
	err_t err;
	int len, sendlen, nrBytesSend;

	err = ERR_OK;

	count = 2 * MAX_TX_SIZE;
	do {
		pHs->buf = mem_malloc(count);
		if (pHs->buf) {
			pHs->buf_len = count;
			break;
		}
		count = count / 2;
	} while (count > 100);

	/* Did we get a send buffer? If not, return immediately. */
	if (pHs->buf == NULL) {
		LWIP_DEBUGF(HTTPD_DEBUG, ("No buff\n"));
		return;
	}
//	if (hs->left == 0)
	LWIP_DEBUGF(HTTPD_DEBUG, ("Sending file\n"));

	do {
		/* Do we have a valid file handle? */
		if (pHs->handle == NULL) {
			//			close_conn(conn, hs);
			return;
		}
		/* Read a block of data from the file. */
		//		LWIP_DEBUGF(HTTPD_DEBUG, ("Trying to read %d bytes.\n", count));
		count = fs_read(pHs->handle, pHs->buf, count);

		LWIP_DEBUGF(HTTPD_DEBUG, ("Read %d bytes\n" , count));

		if (count < 0) {
			/* We reached the end of the file so this request is done */
			LWIP_DEBUGF(HTTPD_DEBUG, ("End of file.\n"));
			fs_close(pHs->handle);
			pHs->handle = NULL;
			return;
		}
		/* Set up to send the block of data we just read */
		//		LWIP_DEBUGF(HTTPD_DEBUG, ("Read %d bytes.\n", count));
		pHs->left = count;
		pHs->file = pHs->buf;

		/* We are not processing an SHTML file so no tag checking is necessary.
		 * Just send the data as we received it from the file.
		 */
		/* We cannot send more data than space available in the send
		 buffer. */
		do {
			sendlen = MAX_TX_SIZE;
			if (sendlen < pHs->left) {
				len = sendlen;
			} else {
				len = pHs->left;
				LWIP_ASSERT("hs->left did not fit into u16_t!\n",(len == pHs->left));
			}
			do {
				printf(":");
#ifdef NETCONS
				err = netconn_write(pConn, pHs->file, len, NETCONN_NOCOPY);
				if (err == ERR_OK)
					nrBytesSend = len;
				else
					nrBytesSend = -1;
#else
				nrBytesSend = send(socket, pHs->file, len, MSG_DONTWAIT); // flags was 0
#endif
				printf(",");

				LWIP_DEBUGF(HTTPD_DEBUG, (" %d bytes send\n",nrBytesSend ));
				if (nrBytesSend == -1) {
					err = ERR_CONN;
					len = 0;
					LWIP_DEBUGF(HTTPD_DEBUG, ("socket error send\n"));
				} else
					len -= nrBytesSend;

				pHs->file += nrBytesSend;
				pHs->left -= nrBytesSend;

			} while (len > 1);

			if (err != ERR_OK)
				pHs->left = 0;

		} while (pHs->left > 0); // send all  bytes from buffer we have just read
	} while (1); // read complete file

	mem_free (pHs->buf);

}

/**
 * reads httpd data
 * checks if file has to be read\n
 * fils http state struct with pointer to file to open or\n
 * executes the cgi script if requested
 *
 * @param[in] pData pointer to ethernet data
 * @param[in] pHs pointer to http_state
 * @param[out] error ERR_OK / !ERR_OK
 */

err_t http_recv(char *pData, struct http_state *pHs) {
	int i, loop, len;
	char *uri;
	struct fs_file *file = NULL;

#ifdef INCLUDE_HTTPD_CGI
	int count;
	char *params;
#endif
	uri = &pData[4];
	LWIP_DEBUGF(HTTPD_DEBUG, ("Request:\n%s\n", pData));
	if (strncmp(pData, "GET ", 4) == 0) {
		/*
		 * We have a GET request. Find the end of the URI by looking for the
		 * HTTP marker. We can't just use strstr to find this since the request
		 * came from an outside source and we can't be sure that it is
		 * correctly formed. We need to make sure that our search is bounded
		 * by the packet length so we do it manually. If we don't find " HTTP",
		 * assume the request is invalid and close the connection.
		 */
		len = strlen(pData);
		for (i = 4; i < (len - 5); i++) {
			if ((pData[i] == ' ') && (pData[i + 1] == 'H')
					&& (pData[i + 2] == 'T') && (pData[i + 3] == 'T')
					&& (pData[i + 4] == 'P')) {
				pData[i] = 0;
				break;
			}
		}
		if (i == (len - 5)) {
			/* We failed to find " HTTP" in the request so assume it is invalid */
			LWIP_DEBUGF(HTTPD_DEBUG, ("Invalid GET request. Closing.\n"));
			//	close_conn(conn, hs);
			return (!ERR_OK);
		}

#ifdef INCLUDE_HTTPD_SSI
		/*
		 * By default, assume we will not be processing server-side-includes
		 * tags
		 */
		pHs->tag_check = false;
#endif
		/*
		 * Have we been asked for the default root file?
		 */
		if ((uri[0] == '/') && (uri[1] == 0)) {
			/*
			 * Try each of the configured default filenames until we find one
			 * that exists.
			 */
			for (loop = 0; loop < NUM_DEFAULT_FILENAMES; loop++) {
				LWIP_DEBUGF(HTTPD_DEBUG,
						("Looking for %s...\n", g_psDefaultFilenames[loop].name));
				file = fs_open((char *) g_psDefaultFilenames[loop].name);
				uri = (char *) g_psDefaultFilenames[loop].name;
				if (file != NULL) {
					LWIP_DEBUGF(HTTPD_DEBUG, ("Opened.\n"));
#ifdef INCLUDE_HTTPD_SSI
					pHs->tag_check = g_psDefaultFilenames[loop].shtml;
#endif
					break;
				}
			}
			if (file == NULL) {
				/* None of the default filenames exist so send back a 404 page */
				file = get_404_file(&uri);
#ifdef INCLUDE_HTTPD_SSI
				pHs->tag_check = false;
#endif
			}
		} else {
			/* No - we've been asked for a specific file. */
#ifdef INCLUDE_HTTPD_CGI
			/* First, isolate the base URI (without any parameters) */

			LWIP_DEBUGF(HTTPD_DEBUG, ("CGI Looking for %s \n", uri));
			params = strchr(uri, '?');
			if (params) {
				*params = '\0';
				params++;
			}
			/* Does the base URI we have isolated correspond to a CGI handler? */
			if (g_iNumCGIs && g_pCGIs) {
				for (i = 0; i < g_iNumCGIs; i++) {
					if (strncmp(uri, g_pCGIs[i].pcCGIName,
							strlen(g_pCGIs[i].pcCGIName)) == 0) {
						LWIP_DEBUGF(HTTPD_DEBUG,
								("CGI %s\n", g_pCGIs[i].pcCGIName));

						/*
						 * We found a CGI that handles this URI so extract the
						 * parameters and call the handler.
						 */
						uri = g_pCGIs[i].pfnCGIHandler(i, params);
						break;
					}
				}
				/* Did we handle this URL as a CGI? If not, reinstate the
				 * original URL and pass it to the file system directly. */
				if (i == g_iNumCGIs) {
					/* Replace the ? marker at the beginning of the parameters */
					if (params) {
						params--;
						*params = '?';
					}
				}
			}
#endif
			LWIP_DEBUGF(HTTPD_DEBUG, ("Opening %s\n", uri));
			file = fs_open(uri);
			if (file == NULL) {

				file = get_404_file(&uri);
			}
#ifdef INCLUDE_HTTPD_SSI
			else
			{
				/*
				 * See if we have been asked for an shtml file and, if so,
				 * enable tag checking.
				 */
				pHs->tag_check = false;
				for(loop = 0; loop < NUM_SHTML_EXTENSIONS; loop++)
				{
					if(strstr(uri, g_pcSSIExtensions[loop]))
					{
						pHs->tag_check = true;
						break;
					}
				}
			}
#endif /* INCLUDE_HTTP_SSI */
		}

		if (file) {
#ifdef INCLUDE_HTTPD_SSI
			pHs->tag_index = 0;
			pHs->tag_state = TAG_NONE;
			pHs->parsed = file->pData;
			pHs->parse_left = file->len;
			pHs->tag_end = file->pData;
#endif
			pHs->handle = file;
			pHs->file = file->data;
			LWIP_ASSERT("File length must be positive!", (file->len >= 0));
			pHs->left = file->len;
			pHs->retries = 0;
		} else {
			pHs->handle = NULL;
			pHs->file = NULL;
			pHs->left = 0;
			pHs->retries = 0;
		}

	} else /// no GET
	{
		return (!ERR_OK);
	}
	return ERR_OK;
}

#ifdef NETCONS

static void close_conn(struct netconn *conn, struct http_state *hs) {
	err_t err;
	LWIP_DEBUGF(HTTPD_DEBUG, ("Closing connection\n"));

	if (hs) {
		if (hs->handle) {
			fs_close(hs->handle);
			hs->handle = NULL;
		}
		if (hs->buf) {
			mem_free(hs->buf);
		}
		mem_free(hs);
	}
	err = netconn_close(conn);
	if (err != ERR_OK) {
		//	LWIP_DEBUGF(HTTPD_DEBUG, ("Error %d closing 0x%08x\n", err, conn));
	}
}

void httpServerServe(struct netconn *conn) {
	struct netbuf *inbuf;
	char *buf;
	u16_t buflen;
	err_t err;
	struct http_state *hs;

	hs = http_accept();
	if (hs != NULL) {
		/* Read the data from the port, blocking if nothing yet there.
		 We assume the request (the part we care about) is in one netbuf */
	//	netconn_set_recvtimeout(conn, HTTPDTIMEOUT);
		err = netconn_recv(conn, &inbuf);
		if (err == ERR_OK) {
			netbuf_data(inbuf, (void**) &buf, &buflen);
			if (http_recv(buf, hs) == ERR_OK)
				send_data(conn, hs);
			netbuf_delete(inbuf);
		}
	}
	close_conn(conn, hs);
	/* Delete the buffer (netconn_recv gives us ownership,
	 so we have to make sure to deallocate the buffer) */
//	netbuf_delete(inbuf);
}

void HTTPServerTask(void *pvParameters) {
	struct netconn *conn, *newconn;
	err_t err = ERR_OK;

	CGI_init();
	while (1) {

		while (!gotIP || !isConnected)
			vTaskDelay(10);

		printf("opening netcon\n");
		conn = netconn_new(NETCONN_TCP);
		netconn_bind(conn, NULL, 80);
	//	netconn_set_recvtimeout(conn, HTTPDTIMEOUT);
		netconn_listen(conn);
	//	if (err == ERR_OK) {
			do {
				err = netconn_accept(conn, &newconn);
				if (err == ERR_OK) {
					httpServerServe(newconn);
					netconn_delete(newconn);
				}
			} while (err == ERR_OK);
	//	}
		netconn_close(conn);
		netconn_delete(conn);
		vTaskDelay(1);

	//	printf("error netcon %d\n", err );

	}
}

#else  // socket based

void HTTPServerTask(void *pvParameters) {
	int socket_fd = 0, accept_fd = 0;
	u32_t addr_size, sent_data;
	int count;

	char receiveBuf[1024];
	struct sockaddr_in sa, isa;

	char *buf;
	u16_t buflen;
	err_t err = false;
	struct http_state *hs;

	struct timeval receiving_timeout;
	receiving_timeout.tv_sec = 5;
	receiving_timeout.tv_usec = 0;

	CGI_init();

	while (1) {
		err = false;

		while (!gotIP || !isConnected)
		vTaskDelay(10);

		memset(&sa, 0, sizeof(struct sockaddr_in));
		sa.sin_family = AF_INET;
		sa.sin_addr.s_addr = htonl(INADDR_ANY);
		sa.sin_port = htons(80);

		printf("opening socket\n");

		socket_fd = socket(PF_INET, SOCK_STREAM, 0);
		if (socket_fd < 0) {
			printf("socket call failed\n");
			err = true;
		}
		if (!err) {
			if (bind(socket_fd, (struct sockaddr *) &sa, sizeof(sa)) == -1) {
				printf("Bind to Port Number 80 failed\n");
				err = true;
			}
		}
		if (!err) {
			printf("Bound to Port Number 80\n");
			listen(socket_fd, 5);
//			if (setsockopt(socket_fd, SOL_SOCKET, SO_RCVTIMEO, &receiving_timeout,
//							sizeof(receiving_timeout)) < 0) {
//				ESP_LOGE(TAG, "... failed to set socket receiving timeout");
//				err = true;
//			}
		}
		while (!err) {
			addr_size = sizeof(isa);
			printf("\nAC");
			accept_fd = accept(socket_fd, (struct sockaddr* ) &isa, &addr_size);
			printf("a ");

			if (accept_fd < 0) {
				printf("accept failed\n");
				err = true;
			}
			if (!err) {
				///	printf("accepted\n");
//				setsockopt(accept_fd, SOL_SOCKET, SO_RCVTIMEO,
//						&receiving_timeout, sizeof(receiving_timeout));
				hs = http_accept();
				if (hs != NULL) {
					printf("R");
					count = recv(accept_fd, receiveBuf, sizeof(receiveBuf), 0);
					printf("r ");
					if (http_recv(receiveBuf, hs) == ERR_OK) {
						printf("S");
						send_data(accept_fd, hs);
						printf("s");
					}
					else err = true;
				}
				close(accept_fd);
			}
			vTaskDelay(1); //allows task to be pre-empted ?
		}
		if (err) {
			close(socket_fd);
			vTaskDelay(1000/portTICK_RATE_MS);
		}
	}
}
#endif

/** @} */
/** @} */

