#include "freertos/FreeRTOS.h"
#include "freertos/portmacro.h"
#include "freertos/event_groups.h"

#include "esp_wifi.h"
#include "esp_system.h"
#include "esp_event.h"
#include "esp_event_loop.h"
#include "esp_log.h"
#include "driver/gpio.h"
#include "nvs_flash.h"


static EventGroupHandle_t wifi_event_group;
const int CONNECTED_BIT = BIT0;

#define TAG "main"

#define mySSID 	"Klepnet"
#define myPWD 	"mypwd"


#define LED_BUILTIN 16
#define delay(ms) (vTaskDelay(ms/portTICK_RATE_MS))

int mountSDCard();
int blinkTime = 100;
int cntr;
bool gotIP;
bool isConnected;
bool doReconnect;


ip4_addr_t myIPAddress;

void thermometerTask(void *pvParameter);
void HTTPServerTask(void *pvParameters);

static void initialise_wifi(void);

static esp_err_t event_handler(void *ctx, system_event_t *event)
{
    switch (event->event_id) {
    case SYSTEM_EVENT_STA_START:
        esp_wifi_connect();
        break;
    case SYSTEM_EVENT_STA_DISCONNECTED:
    	isConnected = false;
    	doReconnect = true;
    	printf("disconnected\n");
//
//    	esp_err_t err =  esp_wifi_connect();
//    	if ( err != ERR_OK)
//    		printf("connect err %d\n ", err);
//        xEventGroupClearBits(wifi_event_group, CONNECTED_BIT);
        break;
    case SYSTEM_EVENT_STA_CONNECTED:
    	printf("connected\n");
    	isConnected = true;
        break;
    case SYSTEM_EVENT_STA_GOT_IP:
		myIPAddress = event->event_info.got_ip.ip_info.ip;

		printf("got ip\n");
		printf("ip: " IPSTR "\n", IP2STR(&event->event_info.got_ip.ip_info.ip));
		printf("netmask: " IPSTR "\n", IP2STR(&event->event_info.got_ip.ip_info.netmask));
		printf("gw: " IPSTR "\n", IP2STR(&event->event_info.got_ip.ip_info.gw));
		printf("\n");
		fflush(stdout);
		gotIP = true;

        xEventGroupSetBits(wifi_event_group, CONNECTED_BIT);
        break;
    case SYSTEM_EVENT_AP_STACONNECTED:
    	gotIP = true;
        ESP_LOGI(TAG, "station:"MACSTR" join,AID=%d\n",
                 MAC2STR(event->event_info.sta_connected.mac),
                 event->event_info.sta_connected.aid);
        xEventGroupSetBits(wifi_event_group, CONNECTED_BIT);
        break;
    case SYSTEM_EVENT_AP_STADISCONNECTED:
    	printf("klp Wifi disconnected \n");

        ESP_LOGI(TAG, "station:"MACSTR"leave,AID=%d\n",
                 MAC2STR(event->event_info.sta_disconnected.mac),
                 event->event_info.sta_disconnected.aid);
        xEventGroupClearBits(wifi_event_group, CONNECTED_BIT);
        gotIP = false;
        break;
    default:
    	printf("event ? %d\n", event->event_id);
        break;
    }
    return ESP_OK;
}


static void initialise_wifi(void) {
	tcpip_adapter_init();
	wifi_event_group = xEventGroupCreate();
	ESP_ERROR_CHECK(esp_event_loop_init(event_handler, NULL));
	wifi_init_config_t cfg = WIFI_INIT_CONFIG_DEFAULT();
	ESP_ERROR_CHECK(esp_wifi_init(&cfg));
//	ESP_ERROR_CHECK(esp_wifi_set_storage(WIFI_STORAGE_RAM));
	ESP_ERROR_CHECK(esp_wifi_set_storage(WIFI_STORAGE_FLASH));
	ESP_ERROR_CHECK(esp_wifi_set_mode(WIFI_MODE_STA));
	wifi_config_t sta_config = { .sta = {
			.ssid = mySSID,
			.password =	myPWD,
			.bssid_set = false } };

	ESP_ERROR_CHECK(esp_wifi_set_config(WIFI_IF_STA, &sta_config));
	ESP_ERROR_CHECK(esp_wifi_start());
}


static void blink(void *pvParameters) {

	while (1) {
		gpio_set_level(LED_BUILTIN, 1);
		vTaskDelay(blinkTime);
		gpio_set_level(LED_BUILTIN, 0);
		vTaskDelay(blinkTime);
		if (gotIP && isConnected)
			blinkTime = 100;
		else
			blinkTime = 20;
	 	if ( doReconnect ){
	 		doReconnect = false;
	    	printf("reconnect\n");
	    	esp_err_t err =  esp_wifi_connect();
	    	if ( err != ESP_OK)
	    		printf("connect err %d\n ", err);
	        xEventGroupClearBits(wifi_event_group, CONNECTED_BIT);
	 	}
	}
}

int app_main(void) {
	gpio_pad_select_gpio(LED_BUILTIN);
	gpio_set_direction(LED_BUILTIN, GPIO_MODE_OUTPUT);

//	if (mountSDCard() == ESP_FAIL)
//		blinkTime = 20;

	ESP_ERROR_CHECK( nvs_flash_init() );
	initialise_wifi();
	xTaskCreate(&blink, "blink", 500, NULL, 1, NULL);
	xTaskCreate(&HTTPServerTask, "http_server", 20000, NULL, 1, NULL);
//	xTaskCreate(&thermometerTask, "thermo", 2000, NULL, 1, NULL);
	return 0;
}
